/* Hue - A Hotline User Editor
 *
 * Copyright (C) 2000 Misha Nasledov <misha@nasledov.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include "hotline.h"
#include <sys/types.h>
#include <curses.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <string.h>
#include "hue.h"
#include <ctype.h>

Hue *hue_with_path(char *path)
{
        Hue *hue;

        for (hue = hue_list; hue; hue = hue->prev) {
                if (hue->path == path)
                        return hue;
        }

        return 0;   
}

Hue *hue_from_win(GtkWidget *window) 
{
	        Hue *hue;

        for (hue = hue_list; hue; hue = hue->prev) {
                if (hue->usermod_window == window)
                        return hue;
        }

        return 0; 
}

void open_user (GtkWidget *widget, gpointer data)
{
	char *path = gtk_file_selection_get_filename (GTK_FILE_SELECTION(data));
	Hue *hue;
	
	hue = hue_new(path);
	if (account_read(path, hue->login, hue->password, hue->name, &hue->acc)) {
		perror(hue->login);
		exit(1);
	}
	gtk_widget_destroy(GTK_WIDGET(data));
	create_usermod_window(hue);
	
}

void open_data (guint callback_data, guint callback_action, GtkWidget *widget)
{
	GtkWidget *filew;
	
	filew = gtk_file_selection_new ("Open");
	
	gtk_signal_connect(GTK_OBJECT(filew), "destroy", (GtkSignalFunc) gtk_widget_destroy, filew);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked", (GtkSignalFunc) gtk_widget_destroy, (gpointer) filew);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked", GTK_SIGNAL_FUNC(open_user), filew);
	
	gtk_widget_show_all(filew);
	gtk_grab_add(filew);
}

void quit_app(guint callback_data, guint callback_action, GtkWidget *widget)
{
	gtk_main_quit();
}

void menu_new(guint callback_data, guint callback_action, GtkWidget *widget)
{
	Hue *hue = hue_new(0);

	create_usermod_window(hue);
}
void prompt_path(Hue *hue, char *type);

void menu_save_as(guint callback_data, guint callback_action, GtkWidget *widget)
{
	prompt_path(hue_front, "Save As...");
}



void hl_code (void *__dst, const void *__src, size_t len)
{
	u_int8_t *dst = (u_int8_t *)__dst, *src = (u_int8_t *)__src;

	for (; len; len--)
		*dst++ = ~*src++;
}

int
account_write (const char *path, const char *login, const char *password, const char *name, const struct hl_access_bits *acc)
{
	int fd, r;
	struct hl_user_data user_data;
	u_int16_t nlen, llen, plen;

	fd = open(path, O_WRONLY|O_CREAT|O_TRUNC, 0600);
	if (fd < 0)
		return errno;

	if ((nlen = strlen(name)) > 134) nlen = 134;
	if ((llen = strlen(login)) > 34) llen = 34;
	if ((plen = strlen(password)) > 32) plen = 32;

	memset(&user_data, 0, sizeof(user_data));
	user_data.magic = htonl(0x00010000);
	user_data.access = *acc;
	user_data.nlen = htons(nlen);
	memcpy(user_data.name, name, nlen);
	user_data.llen = htons(llen);
	memcpy(user_data.login, login, llen);
	user_data.plen = htons(plen);
	hl_code(user_data.password, password, plen);

	if ((r = write(fd, &user_data, 734)) != 734) {
		close(fd);
		return errno;
	}
	fsync(fd);

	close(fd);

	return 0;
}

int
account_read (const char *path, char *login, char *password, char *name, struct hl_access_bits *acc)
{
	struct hl_user_data user_data;
	int fd, r;
	u_int16_t len;

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return errno;
	if ((r = read(fd, &user_data, 734)) != 734) {
		close(fd);
		return errno;
	}
	close(fd);

	if (acc)
		*acc = user_data.access;
	if (name) {
		len = ntohs(user_data.nlen) > 31 ? 31 : ntohs(user_data.nlen);
		memcpy(name, user_data.name, len);
		name[len] = 0;
	}
	if (login) {
		len = ntohs(user_data.llen) > 31 ? 31 : ntohs(user_data.llen);
		memcpy(login, user_data.login, len);
		login[len] = 0;
	}
	if (password) {
		len = ntohs(user_data.plen) > 31 ? 31 : ntohs(user_data.plen);
		hl_code(password, user_data.password, len);
		password[len] = 0;
	}

	return 0;
}

int
test_bit (char *buf, int bitno)
{
	char c, m;
	c = buf[bitno / 8];
	bitno = bitno % 8;
	bitno = 7 - bitno;
	if (!bitno)
		m = 1;
	else {
		m = 2;
		while (--bitno)
			m *= 2;
	}

	return c & m;
}

void hue_delete (Hue *hue)
{
	if (hue->next)
		hue->next->prev = hue->prev;
	if (hue->prev)
		hue->prev->next = hue->next;
	if (hue == hue_list)
		hue_list = hue->prev;
	free(hue);
}

void close_window (GtkWidget *widget, gpointer data)
{
	Hue *hue = 	hue_from_win(widget);

	if(!hue) {
		g_warning("Deleting window with no hue struct...");
	}
	else {
		hue_delete(hue);
	}

	number_open--;
	gtk_widget_destroy(widget);

	if(!number_open) { //no more Hues open, time to quit!
		gtk_main_quit();
	}
}

Hue *hue_new(char *path)
{
	Hue *hue;
	int i;

	hue = malloc(sizeof(Hue));
	hue->next = 0;
	hue->prev = hue_list;
	
	if (hue_list)                                               
		hue_list->next = hue;

	hue->usermod_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(hue->usermod_window, 285, 205);
	gtk_signal_connect(GTK_OBJECT(hue->usermod_window), "delete_event", GTK_SIGNAL_FUNC(close_window), 0);
	hue->pass_entry = gtk_entry_new();
	hue->login_entry = gtk_entry_new();
	hue->name_entry = gtk_entry_new();
	
	for(i = 0; i < ACCESS_NUM; i++) {
		hue->check[i] = gtk_check_button_new_with_label(user_values[i].name);
	}

	memset(hue->name, 0, sizeof(hue->name));
	memset(hue->password, 0, sizeof(hue->password));
	memset(hue->login, 0, sizeof(hue->login)); 
	
	memset(&hue->acc, 0, sizeof(hue->acc)); 
	
	hue->path = g_strdup(path);
	hue_list = hue;

	return hue;
}

void inverse_bit (char *buf, int bitno)
{
        char *p, c, m;
        p = &buf[bitno / 8];
        c = *p;
        bitno = bitno % 8;
        bitno = 7 - bitno;
        if (!bitno)
                m = 1;
        else {
                m = 2;
                while (--bitno)
                        m *= 2;
        }
        if (c & m)
                *p = c & ~m;       
        else
                *p = c | m;
}

void get_path_data(GtkWidget *widget, gpointer data)
{
	Hue *hue = gtk_object_get_data(GTK_OBJECT(widget), "hue");
	char *title;

	hue->path = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(data)));
	gtk_widget_destroy(GTK_WIDGET(data));

	title = g_strdup_printf("User Editor: %s", hue->path);
	gtk_window_set_title(GTK_WINDOW(hue->usermod_window), title);
	g_free(title);

	account_write (hue->path, hue->login, hue->password, hue->name, &hue->acc);
}

void prompt_path(Hue *hue, char *type)
{
	GtkWidget *filew;
	
	filew = gtk_file_selection_new (type);
	
	gtk_signal_connect(GTK_OBJECT(filew), "destroy", (GtkSignalFunc) gtk_widget_destroy, filew);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked", (GtkSignalFunc) gtk_widget_destroy, (gpointer) filew);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked", GTK_SIGNAL_FUNC(get_path_data), filew);
	gtk_object_set_data(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "hue", hue);
	
	gtk_widget_show_all(filew);
	gtk_grab_add(filew);

}

void save_user (GtkWidget *widget, gpointer data)
{
	Hue *hue = data;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(hue->name_entry));
	snprintf(hue->name, 32, "%s", text);
	text = gtk_entry_get_text(GTK_ENTRY(hue->pass_entry));
	snprintf(hue->password, 32, "%s", text);
	text = gtk_entry_get_text(GTK_ENTRY(hue->login_entry));
	snprintf(hue->login, 32, "%s", text);
	
	if(!hue->path) {
		prompt_path(hue, "Save");
		return;
	}

	account_write (hue->path, hue->login, hue->password, hue->name, &hue->acc);
}

void save_user_menu (guint callback_data, guint callback_action, GtkWidget *widget)
{
	char *text;
	
	text = gtk_entry_get_text(GTK_ENTRY(hue_front->name_entry));
	snprintf(hue_front->name, 32, "%s", text);
	text = gtk_entry_get_text(GTK_ENTRY(hue_front->pass_entry));
	snprintf(hue_front->password, 32, "%s", text);
	text = gtk_entry_get_text(GTK_ENTRY(hue_front->login_entry));
	snprintf(hue_front->login, 32, "%s", text);
	
	if(!hue_front->path) {
		prompt_path(hue_front, "Save");
		return;
	}

	account_write (hue_front->path, hue_front->login, hue_front->password, hue_front->name, &hue_front->acc);
}

void toggle_acc (GtkWidget *widget, gpointer data)
{
	Hue *hue = gtk_object_get_data(GTK_OBJECT(widget), "hue");
	inverse_bit((char *)&hue->acc, GPOINTER_TO_INT(data));
}

void init_boxes (Hue *hue)
{
	int i;

	for(i = 0; i < ACCESS_NUM; i++) {
		gtk_toggle_button_set_state((GtkToggleButton *)hue->check[i], test_bit((char *)&hue->acc, user_values[i].bitno)?1:0);
		gtk_signal_connect(GTK_OBJECT(hue->check[i]), "toggled", GTK_SIGNAL_FUNC(toggle_acc), GINT_TO_POINTER(user_values[i].bitno));
		gtk_object_set_data(GTK_OBJECT(hue->check[i]), "hue", hue);
	}
}

void hue_focus (GtkWindow *window, GdkEventFocus *event, gpointer data)
{
	hue_front = data;
}

Hue *create_usermod_window(Hue *hue)
{
	GtkWidget *vbox;
	GtkWidget *name_label;
	GtkWidget *login_label;
	GtkWidget *pass_label;
	GtkWidget *info_table;
	GtkWidget *file_table;
	GtkWidget *general_table;
	GtkWidget *admin_table;
	GtkItemFactory *item_factory;
 	GtkAccelGroup *accel_group;
	GtkWidget *notebook;
	GtkWidget *label;
	char *title;

	number_open++;
	accel_group = gtk_accel_group_new();
	notebook = gtk_notebook_new();
	item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<blah>", accel_group);
	gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
	gtk_accel_group_attach (accel_group, GTK_OBJECT (hue->usermod_window));

	if(!hue->path) {
		gtk_window_set_title(GTK_WINDOW(hue->usermod_window), "User Editor: Untitled");
	}
	else {
		title = g_strdup_printf("User Editor: %s", hue->path);
		gtk_window_set_title(GTK_WINDOW(hue->usermod_window), title);
		g_free(title);
	}
	gtk_window_set_policy(GTK_WINDOW(hue->usermod_window), 1, 1, 0);
	gtk_signal_connect(GTK_OBJECT(hue->usermod_window), "focus-in-event", GTK_SIGNAL_FUNC(hue_focus), hue);

	vbox = gtk_vbox_new(0,0);

	gtk_container_add(GTK_CONTAINER(hue->usermod_window), vbox);
	gtk_box_pack_start(GTK_BOX(vbox),gtk_item_factory_get_widget (item_factory, "<blah>"), 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(vbox), notebook, 0, 0 , 0);
	init_boxes(hue);

	label = gtk_label_new("User Info");
	info_table = gtk_table_new(4, 2, 0);
	gtk_container_add(GTK_CONTAINER(notebook), info_table);
	gtk_entry_set_text(GTK_ENTRY(hue->name_entry), hue->name);
	gtk_entry_set_text(GTK_ENTRY(hue->login_entry), hue->login);
	gtk_entry_set_text(GTK_ENTRY(hue->pass_entry), hue->password);
	name_label = gtk_label_new("Name:");
	login_label = gtk_label_new("Login:");
	pass_label = gtk_label_new("Pass:");
	gtk_table_set_row_spacings(GTK_TABLE(info_table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(info_table), 5);
	gtk_table_attach(GTK_TABLE(info_table), name_label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
	gtk_label_set_justify(GTK_LABEL(name_label), GTK_JUSTIFY_LEFT);               	gtk_misc_set_alignment(GTK_MISC(name_label), 0, 0.5);    
	gtk_container_set_border_width(GTK_CONTAINER(info_table), 5);
	gtk_table_attach(GTK_TABLE(info_table), hue->name_entry, 1, 2, 1, 2, GTK_EXPAND, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(info_table), login_label, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
	gtk_label_set_justify(GTK_LABEL(login_label), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(name_label), 0, 0.5);
	gtk_table_attach(GTK_TABLE(info_table), hue->login_entry, 1, 2, 2, 3, GTK_EXPAND, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(info_table), pass_label, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
	gtk_label_set_justify(GTK_LABEL(pass_label), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(pass_label), 0, 0.5);
	gtk_table_attach(GTK_TABLE(info_table), hue->pass_entry, 1, 2, 3, 4, GTK_EXPAND, 0, 0, 0);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 0), label);

	label = gtk_label_new("Files");
	file_table = gtk_table_new(7, 2, 0);
	gtk_container_add(GTK_CONTAINER(notebook), file_table);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[0], 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[1], 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[20], 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[4], 0, 1, 3, 4, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[2], 0, 1, 4, 5, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[3], 0, 1, 5, 6, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[5], 0, 1, 6, 7, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[8], 1, 2, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[6], 1, 2, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[7], 1, 2, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[23], 1, 2, 3, 4, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[24], 1, 2, 4, 5, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[25], 1, 2, 5, 6, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(file_table), hue->check[26], 1, 2, 6, 7, GTK_FILL, 0, 0, 0);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 1), label);

	label = gtk_label_new("General");
	general_table = gtk_table_new(3, 2, 0);
	gtk_container_add(GTK_CONTAINER(notebook), general_table);
	gtk_table_attach(GTK_TABLE(general_table), hue->check[9], 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(general_table), hue->check[10], 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(general_table), hue->check[21], 0, 1, 2, 3, GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(general_table), hue->check[16], 1, 2, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(general_table), hue->check[15], 1, 2, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(general_table), hue->check[22], 1, 2, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 2), label);

	label = gtk_label_new("Admin");
	admin_table = gtk_table_new(4, 2, 0);
	gtk_container_add(GTK_CONTAINER(notebook), admin_table);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[18], 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[17], 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[19], 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[27], 0, 1, 3, 4, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[13], 1, 2, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[14], 1, 2, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[11], 1, 2, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(admin_table), hue->check[12], 1, 2, 3, 4, GTK_FILL, 0, 0, 0);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 3), label);

	gtk_widget_show_all(hue->usermod_window);
	hue_front = hue;

	return hue;
}
void close_about(GtkWidget *widget, GtkWidget *dialog)
{
	about_open = 0;
	gtk_widget_destroy(dialog);
}

void create_about_window(void)
{
	const char *authors[] = {
		"Misha Nasledov <misha@nasledov.com>",
		"Ryan Nielsen <ran@krazynet.com>", 0 };
	
	GtkWidget *about_box;
	GtkWidget *close;
	GtkWidget *vbox;
	GtkWidget *lbl;
	GtkWidget *hbox;
	GtkWidget *avbox;
	char *lbl_str;
	int i;
	
	if(about_open) return;
	
	about_open = 1;

	about_box = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(about_box), "About Hue");

	close = gtk_button_new_with_label("OK");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(about_box)->action_area), close, 0, 0, 5);

	gtk_signal_connect(GTK_OBJECT(close), "clicked", GTK_SIGNAL_FUNC(close_about), about_box);
	gtk_signal_connect(GTK_OBJECT(about_box), "delete_event", GTK_SIGNAL_FUNC(close_about), about_box);

	vbox = GTK_DIALOG(about_box)->vbox;
	gtk_box_set_spacing(GTK_BOX(vbox), 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	
	lbl_str = g_strdup_printf("%s %s - Hotline User Editor",
PACKAGE, VERSION);
	lbl = gtk_label_new(lbl_str);

	hbox  = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	lbl = gtk_label_new("Copyright (C) 2001 Misha Nasledov");
	hbox = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	
	hbox = gtk_hbox_new(0, 0);
	avbox = gtk_vbox_new(0, 5);
	gtk_box_pack_start(GTK_BOX(hbox), avbox, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	hbox = gtk_hbox_new(0, 0);
	lbl = gtk_label_new("Authors: ");
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(avbox), hbox, 0, 0, 0);
	
	for(i = 0; authors[i]; i++) {
		lbl = gtk_label_new(authors[i]);
		gtk_box_pack_start(GTK_BOX(hbox), lbl, 1, 1, 0);

		hbox = gtk_hbox_new(0, 0);
		gtk_box_pack_start(GTK_BOX(avbox), hbox, 0, 0, 0);
	}
		

	gtk_widget_show_all(about_box);
}


int main (int argc, char *argv[])
{
	Hue *hue;
	char *file = NULL;
	FILE *fd;

	if(argc > 1) {
		file = g_strdup(argv[1]);
	}
	gtk_init(&argc, &argv);

	if(file != NULL) {
		hue = hue_new(file);
		fd = fopen(file, "r");
		if(fd) {
			fclose(fd);
			if (account_read(file, hue->login, hue->password, hue->name, &hue->acc)) {
				perror(hue->login);
				exit(1);
			}
		}
		else {
			//do nothing
		}
		g_free(file);
	}
	else {
		hue = hue_new(0);
	}

	create_usermod_window(hue);
	gtk_main();

	return 0;
}
